
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <strings.h>
#include <errno.h>

extern void panic(const char *fmt, ...);

#include "config.h"
#include "utils.h"
#include "wildmat.h"
#include "log.h"

char *hostname		= NULL;
char *callsign		= NULL;
char *hroute		= NULL;
char *qth		= NULL;
char *importfile	= NULL;
char *importlock	= NULL;
char *exportfile	= NULL;
char *exportlock	= NULL;
char *spooldir		= NULL;
char *innxmit		= NULL;
char *sendmail		= NULL;

static int do_string(char **dest, int argc, char **argv);
static int do_int(int *dest, int argc, char **argv);

#define _CFUNC_ (int (*)(void *dest, int argc, char **argv))

static struct cmd cfg_cmds[] = {
	{ "hostname",		_CFUNC_ do_string,	&hostname	},
	{ "callsign",		_CFUNC_ do_string,	&callsign	},
	{ "hroute",		_CFUNC_ do_string,	&hroute		},
	{ "qth",		_CFUNC_ do_string,	&qth		},
	{ "loglevel",		_CFUNC_ do_int,		&loglevel	},
	{ "logdest",		_CFUNC_ do_int,		&logdest	},
	{ "importfile",		_CFUNC_ do_string,	&importfile	},
	{ "importlock",		_CFUNC_ do_string,	&importlock	},
	{ "exportfile",		_CFUNC_ do_string,	&exportfile	},
	{ "exportlock",		_CFUNC_ do_string,	&exportlock	},
	{ "spooldir",		_CFUNC_ do_string,	&spooldir	},
	{ "innxmit",		_CFUNC_ do_string,	&innxmit	},
	{ "sendmail",		_CFUNC_ do_string,	&sendmail	},
	{ NULL,			NULL,			NULL		}
};

 /* ***************************************************************** */

static int do_string(char **dest, int argc, char **argv)
{
	if (argc < 2)
		return -1;
	*dest = strdup(argv[1]);
	return 0;
}

static int do_int(int *dest, int argc, char **argv)
{
	if (argc < 2)
		return -1;
	*dest = atoi(argv[1]);
	return 0;
}

 /* ***************************************************************** */

/*
 *	Parse c-style escapes (neat to have!)
 */
 
static char *parse_string(char *str)
{
	char *cp = str;
	unsigned long num;

	while (*str != '\0' && *str != '\"') {
		if (*str == '\\') {
			str++;
			switch (*str++) {
			case 'n':
				*cp++ = '\n';
				break;
			case 't':
				*cp++ = '\t';
				break;
			case 'v':
				*cp++ = '\v';
				break;
			case 'b':
				*cp++ = '\b';
				break;
			case 'r':
				*cp++ = '\r';
				break;
			case 'f':
				*cp++ = '\f';
				break;
			case 'a':
				*cp++ = '\007';
				break;
			case '\\':
				*cp++ = '\\';
				break;
			case '\"':
				*cp++ = '\"';
				break;
			case 'x':
				num = strtoul(--str, &str, 16);
				*cp++ = (char) num;
				break;
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
				num = strtoul(--str, &str, 8);
				*cp++ = (char) num;
				break;
			case '\0':
				return NULL;
			default:
				*cp++ = *(str - 1);
				break;
			};
		} else {
			*cp++ = *str++;
		}
	}
	if (*str == '\"')
		str++;		/* skip final quote */
	*cp = '\0';		/* terminate string */
	return str;
}

/*
 *	Parse command line to argv, honoring quotes and such
 */
 
int parse_args(char *argv[],char *cmd)
{
	int ct = 0;
	int quoted;

	while (ct < 31)
	{
		quoted = 0;
		while (*cmd && isspace(*cmd))
			cmd++;
		if (*cmd == 0)
			break;
		if (*cmd == '"') {
			quoted++;
			cmd++;
		}
		argv[ct++] = cmd;
		if (quoted) {
			if ((cmd = parse_string(cmd)) == NULL)
				return 0;
		} else {
			while (*cmd && !isspace(*cmd))
				cmd++;
		}
		if (*cmd)
			*cmd++ = 0;
	}
	argv[ct] = NULL;
	return ct;
}

/*
 *	Find the command from the command table and execute it.
 */
 
int cmdparse(struct cmd *cmds, char *cmdline)
{
	struct cmd *cmdp;
	int argc;
	char *argv[32];

	if ((argc = parse_args(argv, cmdline)) == 0 || *argv[0] == '#')
		return 0;
	strlwr(argv[0]);
	for (cmdp = cmds; cmdp->function != NULL; cmdp++)
		if (strncasecmp(cmdp->name, argv[0], strlen(argv[0])) == 0)
			break;
	if (cmdp->function == NULL)
		return -1;
	return (*cmdp->function)(cmdp->dest, argc, argv);
}

 /* ***************************************************************** */

/*
 *	Read configuration
 */
 
int read_config(void)
{
	FILE *fp;
	char line[LINELEN];
	int ret, n = 0;
	
	if ((fp = fopen(CONF_FILE, "r")) == NULL)
		panic("Cannot open %s", CONF_FILE);
	
	while (fgets(line, LINELEN, fp) != NULL) {
		n++;
		ret = cmdparse(cfg_cmds, line);
		if (ret < 0)
			panic("Syntax error in config file at line %d: %s", n, line);
	}
	fclose(fp);
	
	if (!( (hostname) && (callsign) && (hroute) && (qth) && (importfile)
	   && (importlock) && (exportfile) && (exportlock) ))
		panic("Missing entries in configuration file!");
	
	return 0;
}

/*
 *	Read newsgroup <=> pbbs map file
 */

int read_newsgroup_map(char *fname, int pbbs, struct newsgroup_map **start)
{
	FILE *fp;
	char line[LINELEN], *p, *p1, *p2;
	int n = 0, c;
	int argc;
	char *argv[32];
	struct newsgroup_map *m, **mp;
	
	mp = start;

	if ((fp = fopen(fname, "r")) == NULL)
		panic("Cannot open %s: %s", fname, strerror(errno));

	while (fgets(line, LINELEN, fp) != NULL) {
		n++;

		if ((argc = parse_args(argv, line)) == 0 || *argv[0] == '#')
			continue;
		
		if ((!pbbs) && (argc != 2))
			panic("Wrong number of arguments on line %i in %s. I want two!", n, fname);
		if (argc < 2)
			panic("Wrong number of arguments on line %i in %s. I want at least two!", n, fname);
		
		for (c = 1; c < argc; c++)	/* That's all of c++ i know! */
		if (argv[c][0] != '-') {
			m = malloc(sizeof(struct newsgroup_map));
			memset(m, 0, sizeof(struct newsgroup_map));
			*mp = m;
			mp = &m->next;
		
			m->newsgroup = strdup(argv[0]);
			strlwr(m->newsgroup);
		
			strupr(argv[c]);
			p1 = strtok(argv[c], "<");
			p2 = strtok(NULL, "<");
			
			if (p1) {
				p = strtok(p1, "@");
				if (p) m->to = strdup(p);
				p = strtok(NULL, "@");
				if (p) m->bbs = strdup(p);
			}
			
			if (p2) {
				p = strtok(p2, "@");
				if (p) m->from = strdup(p);
				p = strtok(NULL, "@");
				if (p) m->frombbs = strdup(p);
			}
		}
	}
	fclose(fp);
	return 0;
}

/*
 *	Convert address from type to another using a map file
 */

char *map_address(char *in, char *map)
{
	FILE *f;
	char s[LINELEN];
	char *p, *o;
	int n = 0;
	int argc;
	static char out[ADDRLEN];
	char *argv[32];
	char *repv[9];
	
	if ((f = fopen(map, "r")) == NULL)
		panic("Cannot open %s: %s", map, strerror(errno));

	for (n = 0; n < 9; n++)
		repv[n] = NULL;
	memset(&out, 0, ADDRLEN);

	while (fgets(s, LINELEN, f) != NULL) {
		n++;

		if ((argc = parse_args(argv, s)) == 0 || *argv[0] == '#')
			continue;
			
		if ((argc < 2) || (argc > 3))
			panic("Wrong amount of parameters in %s on line %i!", map, n);
		
		if (!wildmat(in, argv[0], repv))
			continue;
		
		/* Hey, we've got a winner ! */
		
		p = argv[1];
		o = out;
		out[0] = '\0';
		
		for (; (*p); p++)
			if (*p == '$') {
				if (isdigit(*(++p)))
					if (repv[*p - '0'-1] != '\0')
						strcat(o, repv[*p - '0'-1]);
					if(*p == '$')  /* Escaped $ character */
						strcat(o, "$");
				o = out + strlen(out);
			} else
				*o++ = *p;
		
		for (n = 0; (repv[n]); n++)
			free(repv[n]);
		
		fclose(f);
		return out;
	}

	fclose(f);
	return NULL;
}

