#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <syslog.h>
#include <errno.h>

#include "config.h"
#include "rfcutils.h"
#include "pbbs_out.h"
#include "wildmat.h"
#include "crc.h"

/*
 *	Create a PBBS style date/time string: yymmdd/hhmmZ
 */
 
char *pbbstime(const time_t ti)
{
	static char s[13];
	struct tm *t;
	
	t = gmtime(&ti);
	snprintf(s, 13, "%02i%02i%02i/%02i%02iZ", t->tm_year % 100,
		t->tm_mon+1, t->tm_mday, t->tm_hour, t->tm_min);
	
	return s;
}

/*
 *	Create a PBBS mid from a smtp/nntp Message-ID and destination
 *	address
 */
 
char *pbbsmid(char *mid, char *dest)
{
	static char pmid[15];
	char *s;
	char *repv[1];
	                
	if (wildmat(mid, "<*@hamradio>", repv)) {
		strncpy(pmid, repv[0], 15);
		free(repv[0]);
		return pmid;
	} else {
		s = malloc(strlen(mid) + strlen(dest) + 2);
	
		strcpy(s, mid);
		strcat(s, dest);
		
		/*
		strncpy(pmid, crypt(s, "aa"), 15);
		*/
		sprintf(pmid, "%u_MAILGW", calc_crc16(s));
		
		free(s);
	
		return pmid;
	}
}

/*
 *	FBB style file locking
 */
 
struct lhandle *fbblock(char *fname)
{
	struct stat sbuf;
	struct lhandle *lh;
	FILE *lf;
	int i = 0;

	while (stat(fname, &sbuf) == 0) {
		if (i >= SLEEP_TIMES) {
			syslog(LOG_ERR, "Lock file still present after %i seconds. Giving up.", SLEEP_TIME * SLEEP_TIMES);
			return 0;
		}
		i++;
		sleep(SLEEP_TIME);
	}

	if (!(lf = fopen(fname, "w"))) {
		syslog(LOG_ERR, "Couldn't create lock file %s: %s. Giving up.", fname, strerror(errno));
		return 0;
	}
	
	if ((lh = malloc(sizeof(struct lhandle)))) {
		lh->f = lf;
		lh->name = fname;
		return lh;
	} else	/* ouch! */
		return 0;
	
}

int fbbunlock(struct lhandle *lh)
{
	if (fclose(lh->f))
		syslog(LOG_ERR, "Couldn't close lock file %s: %s.", lh->name, strerror(errno));

	if (unlink(lh->name))
		syslog(LOG_ERR, "Couldn't remove lock file %s: %s. Mail transfer will freeze.", lh->name, strerror(errno));

	free(lh);
	return 0;
}

/*
 *	Export to FBB
 */

int export(FILE *f, struct pb_hdr *pb, struct art_hdr *art)
{
	FILE *imf;
	struct lhandle *lh;
	char s[LINELEN];
	int r = 0;
	
	if (!(lh = fbblock(exportlock)))
		return -1;
	
	if (!(imf = fopen(importfile, "a"))) {
		syslog(LOG_ERR, "Couldn't open bbs import file %s: %s. Giving up.", importfile, strerror(errno));
	} else {
		
		fprintf(imf, "%s %s @ %s < %s $%s\n",
			(pb->mtype == MT_B) ? "SB" : "SP", 
			pb->to, pb->tobbs, pb->from, pb->bid);
		fprintf(imf, "%s\n", pb->subj);
		fprintf(imf, "R:%s @:%s [%s] Linux mailgw-%s\n", 
			pbbstime(time(NULL)), hroute, qth, VERSION);
		fprintf(imf, "R:%s @:%s $:%s\n",
			pbbstime(time(NULL)), pb->frombbs, pb->bid);
		
		fprintf(imf, "\n");
		if (art->from)
			fprintf(imf, "T:From: %s\n", art->from);
		if (art->to)
			fprintf(imf, "T:To: %s\n", art->to);
		if (art->newsgroups)
			fprintf(imf, "T:Newsgroups: %s\n", art->newsgroups);
		if (art->mid)
			fprintf(imf, "T:Message-Id: %s\n", art->mid);
		fprintf(imf, "\n");

/*
		fprintf(imf, "\nFrom: %s@%s", pb->from, pb->frombbs);
		if (pb->fromname)
			fprintf(imf, " (%s)", pb->fromname);
		fprintf(imf, "\nTo:   %s@%s", pb->to, pb->tobbs);
		if (pb->toname)
			fprintf(imf, " (%s)", pb->toname);
		fprintf(imf, "\n\n");
*/

		while (fgets(s, LINELEN, f) && (r == 0)) {
			if (!strncasecmp(s, "/ex", 3)) {
				memmove(&s[1], &s[0], LINELEN-1);
				s[0] = ' ';
				s[LINELEN] = '\0';
			}
			if (fputs(s, imf) == EOF) {
				syslog(LOG_ERR, "Couldn't write to bbs import file: %s. Giving up.", strerror(errno));
				r = -1;
			}
		}
		fprintf(imf, "/EX\n");
		if (fclose(imf))
			syslog(LOG_ERR, "Couldn't close bbs import file %s: %s.", importfile, strerror(errno));
	}
	
	fbbunlock(lh);
	
	return r;
}

